"""Compress single cell data using embedding neighborhoods."""

import gc
import numpy as np
import pandas as pd
import scanpy as sc


def _compress_neighborhood(
    adata,
    obs_bio,
    celltype_column,
    additional_groupby_columns,
    measurement_type='gene_expression',
    max_cells_per_type=300,
    avg_neighborhoods=3,
):
    """Compress local neighborhood of a single cell type."""
    # Try something easy first, like k-means
    from sklearn.cluster import KMeans
    from scipy.spatial import ConvexHull

    features = adata.var_names

    groupby_columns = [celltype_column] + list(additional_groupby_columns)

    # Subsample with some regard for cell typing
    # NOTE: this can be debated, but is also designed to get rid of "outlier cells"
    # which are the opposite of an approximation in a sense
    cell_ids = []
    idx = np.zeros(adata.n_obs, bool)
    for _, row in obs_bio.iterrows():
        # Reconstruct indices of focal cells
        idx[:] = True
        for col in groupby_columns:
            idx &= (adata.obs[col] == row[col])

        cell_ids_ct = adata.obs_names[idx]
        ncell = row['cell_count']
        if ncell > max_cells_per_type:
            idx_rand = np.random.choice(range(ncell), size=max_cells_per_type, replace=False)
            cell_ids_ct = cell_ids_ct[idx_rand]
        cell_ids.extend(list(cell_ids_ct))
    adata = adata[cell_ids].copy()

    ##############################################
    # USE AN EXISTING EMBEDDING OR MAKE A NEW ONE
    emb_keys = ['umap', 'tsne']
    for emb_key in emb_keys:
        if f'X_{emb_key}' in adata.obsm:
            break
    else:
        emb_key = 'umap'

        # Log
        sc.pp.log1p(adata)

        # Select features
        sc.pp.highly_variable_genes(adata)
        adata.raw = adata
        adata = adata[:, adata.var.highly_variable]

        # Create embedding, a proxy for cell states broadly
        sc.tl.pca(adata)
        sc.pp.neighbors(adata)
        sc.tl.umap(adata)
        points = adata.obsm[f'X_{emb_key}']

        # Back to all features for storage
        adata = adata.raw.to_adata()
        adata.obsm[f'X_{emb_key}'] = points

        # Back to cptt or equivalent for storage
        adata.X.data = np.expm1(adata.X.data)
    ##############################################

    points = adata.obsm[f'X_{emb_key}']

    # Do a global clustering, ensuring at least 3 cells
    # for each cluster so you can make convex hulls
    if len(groupby_columns) > 1:
        row_order = pd.MultiIndex.from_frame(obs_bio[groupby_columns])
    else:
        row_order = pd.Index(obs_bio[groupby_columns[0]])
    for n_clusters in range(avg_neighborhoods * len(obs_bio), 1, -1):
        kmeans = KMeans(
            n_clusters=n_clusters,
            random_state=0,
            n_init='auto',
        ).fit(points) 
        labels = kmeans.labels_

        # Book keep how many cells of each time are in each cluster
        tmp = adata.obs[groupby_columns].copy()
        tmp['kmeans'] = labels
        tmp['c'] = 1.0
        ncells_per_label = (
                tmp.groupby(['kmeans'] + groupby_columns, observed=False)
                   .size()
                   .unstack(0, fill_value=0)
                   .T
                   )
        del tmp

        # Ensure the order is the same as the averages
        ncells_per_label = ncells_per_label.loc[:, row_order]

        if ncells_per_label.sum(axis=1).min() >= 3:
            break
    else:
        raise ValueError("Cannot cluster neighborhoods")

    n_neis = kmeans.n_clusters
    nei_avg = pd.DataFrame(
            np.zeros((len(features), n_neis), np.float32),
            index=features,
            )
    nei_coords = pd.DataFrame(
            np.zeros((2, n_neis), np.float32),
            index=['x', 'y'],
            )
    convex_hulls = []
    if measurement_type == "gene_expression":
        nei_frac = pd.DataFrame(
                np.zeros((len(features), n_neis), np.float32),
                index=features,
                )
    for i in range(kmeans.n_clusters):
        idx = kmeans.labels_ == i

        # Add the average expression
        nei_avg.iloc[:, i] = np.asarray(adata.X[idx].mean(axis=0))[0]
        # Add the fraction expressing
        if measurement_type == "gene_expression":
            nei_frac.iloc[:, i] = np.asarray((adata.X[idx] > 0).mean(axis=0))[0]

        # Add the coordinates of the center
        points_i = points[idx]
        nei_coords.iloc[:, i] = points_i.mean(axis=0)

        # Add the convex hull
        hull = ConvexHull(points_i)
        convex_hulls.append(points_i[hull.vertices])

    # TODO: Approximate vector field for velocity in this embedding if available.
    # If not, we cannot recount the matrix outselves but if they layers are present,
    # we could recompute it and splash it onto our embedding, THEN recycle the arrow
    # flow algo from mpl to approximate

    # Clean up
    del adata
    gc.collect()

    nei_avg.columns = ncells_per_label.index
    nei_coords.columns = ncells_per_label.index
    if measurement_type == "gene_expression":
        nei_frac.columns = ncells_per_label.index

    neid = {
        'obs_names': nei_avg.index.values,
        'cell_count': ncells_per_label.values,
        'coords_centroid': nei_coords.values,
        'convex_hull': convex_hulls,
        'Xave': nei_avg.values,
    }
    if measurement_type == "gene_expression":
        neid['Xfrac'] = nei_frac.values

    return neid




import numpy as np
import matplotlib.pyplot as plt


def _axis_heuristic(nplots):
    nrows = nplots // 4 + int(bool(nplots % 4))
    ncols = min(nplots, 4)
    return (nrows, ncols)


def embedding(
    accessor,
    features,
    groupby='celltype',
    measurement_type=None,
    key='umap',
    axs=None,
    cmap='viridis_r',
    max_dot_size=100,
    **kwargs,
):
    """Show a neighborhood embedding."""
    if len(features) == 0:
        return

    if axs is None:
        nrows, ncols = _axis_heuristic(len(features))
        fig, axs = plt.subplots(
            nrows, ncols,
            sharex=True, sharey=True,
            figsize=(0.2 + 4 * ncols, 0.2 + 4 * nrows),
        )
        if max(nrows, ncols) == 1:
            axs = [axs]
        elif min(nrows, ncols) > 1:
            axs = axs.ravel()

    adata = accessor.to_adata(
        groupby=groupby,
        neighborhood=True,
        measurement_type=measurement_type,
    )

    hue = adata[:, features].X.copy()
    # TODO: detect log?
    if 'fraction' in adata.layers:
        size = adata[:, features].layers['fraction'].copy()
    else:
        size = hue.copy()

    # Fractions are already between 0 and 1. Scale them remembering that
    # ax.scatter has the funky square thing
    size = max_dot_size * size**2

    centers = adata.obsm[f'X_{key}']

    # Make an embedding for each feature
    for i, feature in enumerate(features):
        ax = axs[i]

        # Scatter the centers
        huei = hue[:, i]
        sizei = size[:, i]
        ax.scatter(
            *centers.T,
            c=huei,
            s=sizei,
            cmap=cmap,
        )

        # Draw the text and convex hulls
        hulls = adata.uns['convex_hulls']
        huemin = huei.min()
        huemax = huei.max() * 1.01 + 1e-10
        for j, hull in enumerate(hulls):
            # Hull
            hueij = huei[j]
            hueij = (hueij - huemin) / (huemax - huemin)
            norm = hueij
            hueij = plt.cm.get_cmap(cmap)(hueij)
            facecolor = tuple(list(hueij)[:3] + [0.7])
            edgecolor = tuple(list(hueij)[:3] + [0.95])
            poly = plt.Polygon(
                hull,
                facecolor=facecolor,
                edgecolor=edgecolor,
            )
            ax.add_patch(poly)

            # Text
            ax.text(
                *centers[j],
                str(j),
                ha='center',
                va='center',
                color='black' if norm < 0.6 else 'white',
            )

        ax.set_title(feature)

    if len(axs) > len(features):
        for i in range(len(features), len(axs)):
            axs[i].set_visible(False)

    fig.tight_layout()

    return axs



import os
import pathlib
import anndata

from .utils import (
    load_config,
    filter_cells,
    normalise_counts,
    correct_annotations,
    approximate_dataset,
    guess_measurement_type,
    guess_normalisation,
    guess_celltype_column,
    guess_celltype_order,
    )
from .io import (
    write_to_h5,
)



class Compressor:
    def __init__(
        self,
        filename=None,
        adata=None,
        output_filename=None,
        celltype_column=None,
        celltype_order=None,
        additional_groupby_columns=tuple(),
        configuration=None,
        include_neighborhood=True,
        ):

        self.filename = filename
        self.adata = adata
        self.output_filename = output_filename
        self.celltype_column = celltype_column
        self.celltype_order = celltype_order
        self.additional_groupby_columns = additional_groupby_columns
        self.configuration = configuration
        self.include_neighborhood = include_neighborhood

    def __call__(self):
        self.prepare()
        self.compress()
        self.store()

    def prepare(self):
        self._validate_constructor()
        self.load()
        self.preprocess()

    def _validate_constructor(self):
        self.configuration = self.configuration or {}
        self.include_neighborhood = bool(self.include_neighborhood)
        if self.output_filename:
            self.output_filename = pathlib.Path(self.output_filename)

    def _guess_annotation_info_if_needed(self):
        if self.celltype_column is None:
            self.celltype_column = guess_celltype_column(self.adata)

        if self.celltype_order is None:
            self.celltype_order = guess_celltype_order(
                self.adata, self.celltype_column,
            )

        if self.additional_groupby_columns is None:
            self.additional_groupby_columns = tuple()

    @property
    def additional_groupby_columns(self):
        return self._additonal_groupby_columns

    @additional_groupby_columns.setter
    def additional_groupby_columns(self, value):
        value = tuple(value)
        if self.celltype_column in value:
            raise ValueError(
                'The cell type column cannot also be an additional column',
            )
        self._additonal_groupby_columns = value

    def load(self):
        if (self.adata is None) and (self.filename is None):
            raise ValueError(
                "Either filename or adata must be specified."
            )

        if self.adata is not None:
            return

        self.adata = anndata.read_h5ad(self.filename)

    def preprocess(self):
        config = self.configuration

        self._guess_annotation_info_if_needed()

        self.adata = filter_cells(self.adata, config)

        self.adata = normalise_counts(
            self.adata,
            config.get("normalisation", guess_normalisation(self.adata)),
            config.get("measurement_type", guess_measurement_type(self.adata)),
        )

        #self.adata = correct_annotations(
        #    self.adata,
        #    self.celltype_column,
        #    config,
        #)

    def compress(self):
        self.approximation = approximate_dataset(
            self.adata,
            self.celltype_column,
            self.additional_groupby_columns,
        )

    def store(self):
        config = self.configuration
        if self.output_filename is not None:
            if self.output_filename.exists():
                os.remove(self.output_filename)
            write_to_h5(
                self.output_filename,
                self.approximation,
            )



def _infer_dtype(dtype):
    if str(dtype) == 'object':
        return 'S'
    if str(dtype).startswith('S'):
        return 'S'
    if str(dtype).startswith('U'):
        return 'S'
    # FIXME: improve this
    if str(dtype) == 'category':
        return 'S'

    return dtype




'''
Utility functions for the compression
'''
import os
import gc
import pathlib
import numpy as np
import pandas as pd
import h5py
import scanpy as sc

from scquill.utils.types import _infer_dtype


def write_to_h5(
        fn_out,
        compressed_atlas,
        compression=22,
        quantisation="chromatin_accessibility",
        #chunked=True,
        ):
    '''Store compressed atlas into h5 file.

    Args:
        fn_out: The h5 file with the compressed atlas.
        compressed_atlas: The dict with the result.
        groupby_order: The order of groups, typically cell types with potentially additionaal
            metadata such as time point and disease, that constitute the biological unit of
            approximation.
        quantisation: If not None, average measurement is quantised with these bins.
        compression: Use zstd compression of the data arrays (avg and frac). Levels are 1-22,
            whereas 0 or False means no compression. No performace decrease is observed.
    '''
    add_kwargs = {}

    # Optional zstd compression using hdf5plugin
    if compression:
        try:
            import hdf5plugin
            # NOTE: decompressing zstd is equally fast no matter how much compression.
            # As for compression speed, levels 1-19 are normal, 20-22 "ultra".
            # A quick runtime test shows *faster* access for clevel=22 than clevel=3,
            # while the file size is around 10% smaller. Compression speed is significantly
            # slower, but (i) still somewhat faster than actually averaging the data and
            # (ii) compresses whole human RNA+ATAC is less than 1 minute. That's nothing
            # considering these approximations do not change that often.
            comp_kwargs = hdf5plugin.Zstd(clevel=compression)
        except ImportError:
            # TODO: throw some warning
            compression = False
            comp_kwargs = {}
    else:
        comp_kwargs = {}

    with h5py.File(fn_out, 'a') as h5_data:
        # Version
        import scquill
        h5_data.attrs['scquill_version'] = scquill.__version__

        # Umbrella group for measurments (not metadata)
        me_all = h5_data.create_group('measurements')

        for measurement_type, approximation_dict_mt in compressed_atlas.items():

            # TODO: move the quantisation bit somewhere else
            # Data can be quantised for further compression (typically ATAC-Seq)
            if (quantisation == True) or (quantisation == measurement_type):
                if measurement_type == "chromatin_accessibility":
                    # NOTE: tried using quantiles for this, but they are really messy
                    # and subject to aliasing effects. 8 bits are more than enough for
                    # most biological questions given the noise in the data
                    qbits = 8
                    bins = np.array([-0.001, 1e-8] + np.logspace(-4, 0, 2**qbits - 1).tolist()[:-1] + [1.1])
                    # bin "centers", aka data quantisation
                elif measurement_type == "gene_expression":
                    # Counts per ten thousand quantisation
                    qbits = 16
                    bins = np.array([-0.001, 1e-8] + np.logspace(-2, 4, 2**qbits - 1).tolist()[:-1] + [1.1e4])
                else:
                    raise ValueError(f"Quantisation for {measurement_type} not set.")

                quantisation_array = [0] + np.sqrt(bins[1:-2] * bins[2:-1]).tolist() + [1]

                qbytes = qbits // 8
                # Add a byte if the quantisation is not optimal
                if qbits not in (8, 16, 32, 64):
                    qbytes += 1
                avg_dtype = f"u{qbytes}"
                quantisation = True
            else:
                avg_dtype = "f4"
                quantisation = False

            features = approximation_dict_mt['var_names'].tolist()

            # Subgroup for a specific measurement type (multi-omics will have multiple)
            me = me_all.create_group(measurement_type)
    
            me.attrs['compression'] = compression

            me.create_dataset('var_names', data=np.array(features).astype('S'))
            if quantisation:
                me.create_dataset('quantisation', data=np.array(quantisation_array).astype('f4'))

            # Number of cells
            ncells = approximation_dict_mt['obs']['cell_count']
            me.create_dataset(
                'cell_count', data=ncells.values, dtype='i8')

            # Grouping information and metadata
            obs = approximation_dict_mt['obs']
            groupby_columns = obs.columns[:-1]
            groupby = me.create_group("groupby")
            groupby.attrs['n_levels'] = len(groupby_columns)
            groupby_names = groupby.create_group("names")
            groupby_types = groupby.create_group("dtypes")
            for i, col in enumerate(groupby_columns):
                groupby_names.attrs[str(i)] = col
                groupby_types.attrs[str(i)] = str(obs[col].dtype)
            meta = me.create_group('obs')
            for i, column in enumerate(obs.columns):
                dtype_store = _infer_dtype(obs[column].dtype)
                meta.create_dataset(
                    column,
                    data=obs[column].values.astype(dtype_store),
                )

            # Average in a cell type
            avg = approximation_dict_mt['Xave']
            if quantisation:
                # pd.cut wants one dimensional arrays so we ravel -> cut -> reshape
                avg_vals = (pd.cut(avg.ravel(), bins=bins, labels=False)
                            .reshape(avg.shape)
                            .astype(avg_dtype))

            # TODO: manual chunking might increase performance a bit, the data is
            # typically accessed only vertically (each feature its own island)
            #if chunked:
            #    # Chunk each feature on its own: this is perfect for ATAC-Seq 
            #    add_kwargs['chunks'] = (1, len(features))

            # Groups (cell types w/ or w/o additional metadata)
            obs_names = approximation_dict_mt['obs_names']
            me.create_dataset(
                'obs_names', data=obs_names.values.astype('S'),
            )

            # Average and fraction detected
            me.create_dataset(
                'average', data=avg.T, dtype=avg_dtype,
                **add_kwargs,
                **comp_kwargs,
            )
            if measurement_type == 'gene_expression':
                # Fraction detected in a cell type
                frac = approximation_dict_mt['Xfrac']
                me.create_dataset(
                    'fraction', data=frac.T, dtype='f4',
                    **add_kwargs,
                    **comp_kwargs,
                )

            # Local neighborhoods
            neid = approximation_dict_mt['neighborhood']
            neigroup = me.create_group('neighborhood')
            ncells = neid['cell_count']
            neigroup.create_dataset(
                'cell_count', data=ncells, dtype='i8')

            avg = neid['Xave']
            if quantisation:
                # pd.cut wants one dimensional arrays so we ravel -> cut -> reshape
                avg_vals = (pd.cut(avg.ravel(), bins=bins, labels=False)
                            .reshape(avg.shape)
                            .astype(avg_dtype))
            # NOTE: these are just "neighborhood 0" etc. so no need for fancy joining
            neigroup.create_dataset(
                'obs_names', data=neid['obs_names'].astype('S'))
            neigroup.create_dataset(
                'average', data=avg.T, dtype=avg_dtype,
                **add_kwargs,
                **comp_kwargs,
            )
            if measurement_type == 'gene_expression':
                # Fraction detected in a cell type
                frac = neid['Xfrac']
                neigroup.create_dataset(
                    'fraction', data=frac.T, dtype='f4',
                    **add_kwargs,
                    **comp_kwargs,
                )

            # Centroid coordinates
            coords_centroids = neid['coords_centroid']
            neigroup.create_dataset(
                'coords_centroid',
                data=coords_centroids.T, dtype=avg_dtype,
                **add_kwargs,
                **comp_kwargs,
            )

            # Convex hulls
            convex_hulls = neid['convex_hull']
            hullgroup = neigroup.create_group('convex_hull')
            for ih, hull in enumerate(convex_hulls):
                hullgroup.create_dataset(
                    str(ih), data=hull, dtype='f4',
                    **add_kwargs,
                    **comp_kwargs,
                )




import importlib.metadata

__version__ = importlib.metadata.version("scquill")

from .compressor import Compressor
from .approximation import Approximation
from .utils import coarse_grain_anndata

import scquill.pl as pl


__all__ = (
    "__version__",
    "Compressor",
    "Approximation",
    "pl",
)



import re
import numpy as np
import anndata


def guess_normalisation(adata):
    """Guess normalisation of the single cell data"""
    if (adata.n_obs == 0):
        raise ValueError("Single cell data set must have at least one observation/cell.""")

    sum0 = int(adata.X[0].sum())
    if sum0 == 10000:
        return 'cptt'
    if sum0 == 1000000:
        return 'cpm'

    is_integer = (np.floor(adata.X.data[:30]) == np.ceil(adata.X.data[:30])).all()
    if is_integer:
        return 'raw'

    if adata.X.max() > 50:
        raise ValueError("Could not guess normalisation of this data set")

    Xe = np.expm1(adata.X)
    adatae = anndata.AnnData(X=Xe)
    norm = guess_normalisation(adatae)
    if norm == 'cpm':
        return 'cpm+logp1'
    if norm == 'cptt':
        return 'cptt+logp1'
    return 'log'


def guess_measurement_type(adata):
    """Guess measurement type (gene expression, chromatin accessibility)"""
    var0 = adata.var_names[0]
    
    # ATAC-Seq peak patterns:
    # chr1-900-1000
    # Chr10-984-8222
    # 1-8894-2299
    # chr5_3999_90000
    pattern = '^(?:chr|Chr|)[0-9]+[_-][0-9]+[_-][0-9]+$'

    if re.findall(pattern, var0):
        return 'chromatin_accessibility'

    return 'gene_expression'


def guess_celltype_column(adata):
    guesses = [
        'celltype',
        'cell_type',
        'cell-type',
        'CellType',
        'cellType',
        'Cell_Type',
        'cell_annotation',
        'CellAnnotation',
        'cellAnnotation',
        'annotation',
        'cell_ontology_class',
        'free_annotation',
        'broad_type',
        'type',
    ]

    cols = adata.obs.columns

    # Exact matches
    for guess in guesses:
        if guess in cols:
            return guess

    # Exact substring
    for guess in guesses:
        for col in cols:
            if guess in col:
                return col

    raise ValueError("Cannot guess celltype column")


def guess_celltype_order(adata, celltype_column):
    """Guess the order of cell types"""
    # TODO: try to come up with something better than this...
    unique = adata.obs[celltype_column].value_counts().index
    return list(unique)



'''
Utility functions for the compression
'''
import gc
import numpy as np
import pandas as pd
import scanpy as sc


from .biological import _compress_biological
from .neighborhood import _compress_neighborhood


def approximate_dataset(
    adata,
    celltype_column,
    additional_groupby_columns=tuple(),
    measurement_type="gene_expression",
):
    """Compress atlas for one tissue after data is clean, normalised, and reannotated."""

    # Celltype averages
    biod = _compress_biological(
        adata,
        celltype_column,
        additional_groupby_columns,
        measurement_type,
    )

    # Local neighborhoods
    neid = _compress_neighborhood(
        adata,
        biod['obs'],
        celltype_column,
        additional_groupby_columns,
        measurement_type,
    )

    features = adata.var_names
    result = {
        'var_names': features,
        'obs': biod['obs'],
        'obs_names': biod['obs_names'],
        'Xave': biod['Xave'],
    }
    if measurement_type == "gene_expression":
        result['Xfrac'] = biod['Xfrac']
    result['neighborhood'] = neid

    compressed_atlas = {
        measurement_type: result,
    }

    return compressed_atlas






'''
Utility functions for the compression
'''
import os
import gc
import pathlib
import yaml
import gzip
import numpy as np
import pandas as pd
import h5py

import scanpy as sc

from .config import load_config
from .preprocess import (
    filter_cells,
    normalise_counts,
    correct_annotations,
)
from .compress import (
    approximate_dataset,
)
from .heuristics import (
    guess_normalisation,
    guess_measurement_type,
    guess_celltype_column,
    guess_celltype_order,
)
from .coarse_grain import (
    coarse_grain_anndata,
)



"""Read from files."""
import pandas as pd
import anndata

from scquill.utils.types import _infer_dtype


def read_h5_to_anndata(
    h5_data,
    neighborhood,
    measurement_type,
):
    """Get an AnnData object in which each observation is an average."""
    if measurement_type not in h5_data["measurements"]:
        raise KeyError("Measurement type not found: {measurement_type}")

    me = h5_data["measurements"][measurement_type]
    compression = me.attrs.get("compression", None)

    if compression:
        try:
            import hdf5plugin
        except ImportError:
            raise ImportError(
                'You need the "hdf5plugin" package to decompress this approximation. You can install it e.g. via pip install hdf5plugin.'
            )

    var_names = me["var_names"].asstr()[:]

    if "quantisation" in me:
        quantisation = me["quantisation"][:]

    groupby_names = []
    groupby_dtypes = []
    n_levels = me["groupby"].attrs["n_levels"]
    for i in range(n_levels):
        groupby_names.append(me["groupby"]["names"].attrs[str(i)])
        groupby_dtypes.append(me["groupby"]["dtypes"].attrs[str(i)])
    groupby = "\t".join(groupby_names)

    if neighborhood:
        neigroup = me["neighborhood"]
        Xave = neigroup["average"][:]
        if "quantisation" in me:
            Xave = quantisation[Xave]

        groupby_order = me["obs_names"].asstr()[:]
        obs_names = neigroup["obs_names"].asstr()[:]
        ncells = neigroup["cell_count"][:]
        coords_centroid = neigroup["coords_centroid"][:]
        convex_hulls = []
        for ih in range(len(coords_centroid)):
            convex_hulls.append(neigroup["convex_hull"][str(ih)][:])

        if measurement_type == "gene_expression":
            Xfrac = neigroup["fraction"][:]
            adata = anndata.AnnData(
                X=Xave,
                layers={
                    "average": Xave,
                    "fraction": Xfrac,
                },
            )
        else:
            adata = anndata.AnnData(X=Xave)

        adata.obsm["X_ncells"] = ncells
        adata.obsm["X_umap"] = coords_centroid
        adata.uns["convex_hulls"] = convex_hulls

        adata.obs_names = pd.Index(obs_names, name="neighborhoods")
        adata.var_names = pd.Index(var_names, name="features")

    else:
        Xave = me["average"][:]
        if "quantisation" in me:
            Xave = quantisation[Xave]

        if measurement_type == "gene_expression":
            Xfrac = me["fraction"][:]
        obs_names = me["obs_names"].asstr()[:]
        # Add obs metadata
        obs = pd.DataFrame([], index=obs_names)
        for column, dtype in zip(groupby_names, groupby_dtypes):
            if _infer_dtype(dtype) == "S":
                obs[column] = me["obs"][column].asstr()[:]
            else:
                obs[column] = me["obs"][column][:]
        obs["cell_count"] = me['cell_count'][:]

        if measurement_type == "gene_expression":
            adata = anndata.AnnData(
                X=Xave,
                obs=obs,
                layers={
                    "average": Xave,
                    "fraction": Xfrac,
                },
            )
        else:
            adata = anndata.AnnData(
                X=Xave,
                obs=obs,
            )

        adata.var_names = pd.Index(var_names, name="features")
        adata.obs_names = pd.Index(obs_names, name=groupby)

    adata.uns["approximation_groupby"] = {
        "names": groupby_names,
        "dtypes": groupby_dtypes,
    }
    if neighborhood:
        adata.uns["approximation_groupby"]["order"] = groupby_order
        adata.uns["approximation_groupby"]["cell_count"] = me['cell_count'][:]

    return adata




import numpy as np
import matplotlib.pyplot as plt

from .dotplot import dotplot
from .neighborhood_composition import neighborhood_composition


def neighborhoodplot(
    accessor,
    features,
    groupby='celltype',
    measurement_type=None,
    axs=None,
    cmap='viridis_r',
    max_dot_size=100,
    **kwargs,
):
    """Composition of all neighborhoods."""

    if axs is None:
        adata = accessor.to_adata(
            groupby=groupby,
            neighborhood=True,
            measurement_type=measurement_type,
        )
        nnei = adata.n_obs
        nct = len(adata.uns['celltypes'])
        fig, axs = plt.subplots(
            1, 2,
            gridspec_kw=dict(
                width_ratios=[0.3 + len(features), nct],
                ),
            sharey=True,
            figsize=(0.5 + 0.3 * (len(features) + nct), 0.5 + 0.3 * nnei),
            )


    dotplot(
        accessor,
        features,
        groupby=groupby,
        measurement_type=measurement_type,
        neighborhood=True,
        ax=axs[0],
        max_dot_size=max_dot_size,
        cmap=cmap,
        **kwargs,
    )
    neighborhood_composition(
        accessor,
        groupby=groupby,
        measurement_type=measurement_type,
        ax=axs[1],
        cmap=cmap,
        **kwargs,
    )
    axs[0].set_ylabel(axs[1].get_ylabel())
    axs[1].set_ylabel('')
    fig.tight_layout()

    return axs



import numpy as np
import matplotlib.pyplot as plt


def dotplot(
    accessor,
    features,
    groupby='celltype',
    neighborhood=False,
    measurement_type=None,
    ax=None,
    cmap='viridis_r',
    max_dot_size=100,
    **kwargs,
):
    """Make a dot plot of features x groups."""
    if len(features) == 0:
        return

    if ax is None:
        ax = plt.gca()

    adata = accessor.to_adata(
        groupby=groupby,
        neighborhood=neighborhood,
        measurement_type=measurement_type,
    )

    hue = adata[:, features].X.copy()
    # TODO: detect log?
    if 'fraction' in adata.layers:
        size = adata[:, features].layers['fraction'].copy()
    else:
        size = hue.copy()

    # Fractions are already between 0 and 1. Scale them remembering that
    # ax.scatter has the funky square thing
    size = max_dot_size * size**2

    # Create x and y coordinates in a mesh
    ngroups, nfeatures = hue.shape
    xs, ys = np.meshgrid(np.arange(nfeatures), np.arange(ngroups))

    ax.scatter(
        xs.ravel(),
        ys.ravel(),
        s=size.ravel(),
        c=hue.ravel(),
        cmap=cmap,
    )
    ax.set_xlim(-0.5, nfeatures - 0.5)
    ax.set_xticks(np.arange(nfeatures))
    ax.set_yticks(np.arange(ngroups))
    ax.set_xticklabels(features, rotation=90)
    ax.set_yticklabels(adata.obs_names)
    ax.figure.tight_layout()

    return ax



import gc
import numpy as np
import scanpy as sc


def filter_cells(adata, config_mt):
    """Filter cells according to some parameter dictionary."""
    filter_dict = config_mt.get("filter_cells", {})

    if len(filter_dict) == 0:
        return adata

    ncells_orig = adata.shape[0]

    if "min_cells_per_type" in filter_dict:
        nmin = filter_dict["min_cells_per_type"]

        column = config_mt["cell_annotations"]["column"]
        ct_counts = adata.obs[column].value_counts()
        ct_abundant = ct_counts.index[ct_counts >= nmin]
        adata = adata[adata.obs[column].isin(ct_abundant)]

    if "unannotated" in filter_dict:
        unanno_values = filter_dict["unannotated"]
        if isinstance(unanno_values, str):
            unanno_values = [unanno_values]

        column = config_mt["cell_annotations"]["column"]
        # If a list is given, take the first one you find or fail
        if not isinstance(column, str):
            columns = column
            for column in columns:
                if column in adata.obs.columns:
                    break
            else:
                raise ValueError(
                    f"None of the cell annotation columns found: {columns}",
                )
        adata = adata[~adata.obs[column].isin(unanno_values)]

    ncells_new = adata.shape[0]

    if ncells_new < ncells_orig:
        delta = ncells_orig - ncells_new
        print(f'Filtered out {delta} cells, originally {ncells_orig} cells, {ncells_new} remaining')

    return adata


def normalise_counts(adata_tissue, input_normalisation, measurement_type="gene_expression"):
    """Normalise counts no matter what the input normalisation is."""
    if measurement_type == "gene_expression":
        if input_normalisation not in (
                "cptt", "raw", "cpm", "cpm+log", "cptt+log", "to-raw", "to-raw+cptt+log"):
            raise ValueError("Input normalisation not recognised: {input_normalisation}")

        if input_normalisation in ("to-raw", "to-raw+cptt+log"):
            adata_tissue = adata_tissue.raw.to_adata()

        if input_normalisation in ("cpm+log", "cptt+log", "to-raw+cptt+log"):
            adata_tissue.X = np.expm1(adata_tissue.X)

        if input_normalisation in ("raw", "cpm", "cpm+log", "to-raw"):
            sc.pp.normalize_total(
                adata_tissue,
                target_sum=1e4,
                key_added='coverage',
            )

        return adata_tissue

    elif measurement_type == "chromatin_accessibility":
        if input_normalisation not in ("binary", "to-binary"):
            raise ValueError("Input normalisation not recognised: {input_normalisation}")

        if input_normalisation == "to-binary":
            adata_tissue.X.data[:] = 1

        return adata_tissue

    raise ValueError("measurement type not recognised")


def subannotate(adata,
                species, annotation,
                markers,
                bad_prefixes=None,
                verbose=True,
                trash_unknown=True,
                skip_subannotation=False):
    '''This function subannotates a coarse annotation from an atlasi.

    This is ad-hoc, but that's ok for now. Examples are 'lymphocyte', which is
    a useless annotation unless you know what kind of lymphocytes these are, or
    if it's a mixed bag.
    '''
    # If skipping, return list of empty annotations - basically blacklisting
    if skip_subannotation:
        return [""] * adata.shape[0]

    if bad_prefixes is None:
        bad_prefixes = []

    markersi = markers.get(annotation, None)
    if markersi is None:
        raise ValueError(
            f'Cannot subannotate without markers for {species}, {annotation}')

    adata = adata.copy()
    sc.pp.log1p(adata)

    genes, celltypes = [], []
    for celltype, markers_ct in markersi.items():
        celltypes.append(celltype)
        for gene in markers_ct:
            if gene in adata.var_names:
                genes.append(gene)
            elif verbose:
                print('Missing gene:', gene)

    adatam = adata[:, genes].copy()

    # No need for PCA because the number of genes is small

    # Get neighbors
    sc.pp.neighbors(adatam)

    # Get communities
    sc.tl.leiden(adatam)

    adata.obs['subleiden'] = adatam.obs['leiden']
    sc.tl.rank_genes_groups(
        adata,
        'subleiden',
        method='t-test_overestim_var',
    )
    top_marker = pd.DataFrame(adata.uns['rank_genes_groups']['names']).head(2)

    subannos = {}
    for cluster, genestop in top_marker.items():
        found = False
        for gene in genestop:
            if found:
                break
            found_bad_prefix = False
            for bad_pfx in bad_prefixes:
                if gene.startswith(bad_pfx):
                    found_bad_prefix = True
                    break
            if found_bad_prefix:
                subannos[cluster] = ''
                continue
            for celltype, markers_ct in markersi.items():
                if gene in markers_ct:
                    subannos[cluster] = celltype
                    found = True
                    break
            else:
                # FIXME: trash clusters with unknown markers for now
                if not trash_unknown:
                    import ipdb; ipdb.set_trace()
                    raise ValueError('Marker not found:', gene)
                else:
                    subannos[cluster] = ''
        if not found:
            subannos[cluster] = ''

    new_annotations = adata.obs['subleiden'].map(subannos)

    return new_annotations


def correct_annotations(
    adata,
    column,
    species,
    tissue,
    rename_dict,
    require_subannotation,
    blacklist=None,
    subannotation_kwargs=None,
):
    '''Correct cell types in each tissue according to known dict'''
    # If a list is given, take the first one you find or fail
    if not isinstance(column, str):
        columns = column
        for column in columns:
            if column in adata.obs.columns:
                break
        else:
            raise ValueError(
                f"None of the cell annotation columns found: {columns}",
            )

    # Ignore cells with NaN in the cell.type column
    idx = adata.obs[column].isin(
            adata.obs[column].value_counts().index)
    adata = adata[idx].copy()

    gc.collect()

    adata.obs[column + '_lowercase'] = adata.obs[column].str.lower()

    if blacklist is None:
        blacklist = {}
    if subannotation_kwargs is None:
        subannotation_kwargs = {}

    celltypes_new = np.asarray(adata.obs[column + '_lowercase']).copy()

    # Exclude blacklisted
    if tissue in blacklist:
        for ctraw in blacklist[tissue]:
            celltypes_new[celltypes_new == ctraw] = ''

    # Rename according to standard dict
    if 'cell_types' in rename_dict:
        for ctraw, celltype in rename_dict['cell_types'].items():
            # one can use brain:neuron for renaming in specific tissues only
            if isinstance(ctraw, str) and (':' not in ctraw):
                celltypes_new[celltypes_new == ctraw] = celltype
            else:
                # Organ-specific renames
                if isinstance(ctraw, str) and ':' in ctraw:
                    organraw, ctraw = ctraw.split(':')
                else:
                    organraw, ctraw = ctraw
                if organraw == tissue:
                    celltypes_new[celltypes_new == ctraw] = celltype

    ct_found = np.unique(celltypes_new)

    # In some data sets, some unnotated clusters are denoted by a digit
    for ctraw in ct_found:
        if ctraw.isdigit():
            celltypes_new[celltypes_new == ctraw] = ''
    ct_found = np.unique(celltypes_new)

    # Look for coarse annotations
    ctnew_list = set(celltypes_new)
    for celltype in ctnew_list:
        if celltype in require_subannotation:
            idx = celltypes_new == celltype
            adata_coarse_type = adata[idx]
            print(f'Subannotating {celltype}')
            subannotations = subannotate(
                adata_coarse_type, species, celltype,
                **subannotation_kwargs,
            )

            # Ignore reclustering into already existing types, we have enough
            for subanno in subannotations:
                if subanno in ct_found:
                    subannotations[subannotations == subanno] = ''
            print('Subannotation done')

            celltypes_new[idx] = subannotations

    adata.obs['cellType'] = celltypes_new

    # Eliminate cell types with less than 3 cells
    ncells = adata.obs['cellType'].value_counts()
    rare_celltypes = ncells.index[ncells < 3]
    adata.obs.loc[adata.obs['cellType'].isin(rare_celltypes), 'cellType'] = ''

    # Correction might declare some cells as untyped/low quality
    # they have an empty string instead of an actual annotation
    if (adata.obs['cellType'] == '').sum() > 0:
        idx = adata.obs['cellType'] != ''
        adata= adata[idx]

    return adata



import h5py
import pandas as pd
import anndata

from .utils.coarse_grain import coarse_grain_anndata
from .utils.types import _infer_dtype
from .io import read_h5_to_anndata


class Approximation:
    """Access single cell approximations."""

    def __init__(
        self,
    ):
        self._adata_dict = {}

    @classmethod
    def read_h5(
        cls,
        filename,
    ):
        """Lazy reader of approximation from file."""
        self = cls()
        self.approximation_dict = None
        self.filename = filename
        return self

    @classmethod
    def read_approximation_dict(
        cls,
        approximation_dict,
    ):
        """Lazy reader of approximation from dict."""
        self = cls()
        self.filename = None
        self.approximation_dict = approximation_dict
        return self

    def _infer_measurement_type(self, h5_data):
        measurement_types = list(h5_data["measurements"])
        if len(measurement_types) == 1:
            return measurement_types[0]
        elif len(measurement_types) == 0:
            raise KeyError("No measurements found in this approximation")
        else:
            raise KeyError("Multiple measurement types found: {measurement_types}")

    def _to_anndata(
        self,
        neighborhood=False,
        measurement_type=None,
    ):
        if self.approximation_dict:
            adata = self._appdict_to_anndata(
                neighborhood=neighborhood,
                measurement_type=measurement_type,
            )
        else:
            adata = self._h5file_to_anndata(
                neighborhood=neighborhood,
                measurement_type=measurement_type,
            )
        self._adata_dict[(measurement_type, neighborhood)] = adata

    def _h5file_to_anndata(
        self,
        neighborhood,
        measurement_type,
    ):
        with h5py.File(self.filename) as h5_data:
            if measurement_type is None:
                measurement_type = self._infer_measurement_type(h5_data)
            adata = read_h5_to_anndata(
                h5_data,
                neighborhood,
                measurement_type,
            )
        return adata

    def _appdict_to_anndata(
        neighborhood,
        measurement_type,
    ):
        compressed_atlas = self.approximation_dict

        if measurement_type is None:
            if len(compressed_atlas.keys()) == 1:
                measurement_type = list(compressed_atlas.keys())[0]
            else:
                raise ValueError(
                    "Multiple measurement types detected, which one would you like to look at?",
                )

        resd = compressed_atlas[measurement_type]
        var_names = resd["features"]

        if neighborhood:
            neid = resd["neighborhood"]
            Xave = neid["avg"].values
            obs_names = neid["avg"].index.values

            if measurement_type == "gene_expression":
                Xfrac = neid["frac"].values
                adata = anndata.AnnData(
                    X=Xave,
                    layers={
                        "average": Xave,
                        "fraction": Xfrac,
                    },
                )
            else:
                adata = anndata.AnnData(X=Xave)

            adata.obs["cell_count"] = neid["ncells"]
            adata.obsm["X_ncells"] = neid["ncells"]
            adata.obsm["X_umap"] = neid["coords_centroid"]
            adata.uns["convex_hulls"] = neid["convex_hull"]
            adata.obs_names = pd.Index(obs_names, name="neighborhoods")
            adata.var_names = pd.Index(var_names, name="features")

        else:
            Xave = resd["avg"].values
            if measurement_type == "gene_expression":
                Xfrac = resd["frac"].values
                adata = anndata.AnnData(
                    X=Xave,
                    layers={
                        "average": Xave,
                        "fraction": Xfrac,
                    },
                )
            else:
                adata = anndata.AnnData(
                    X=Xave,
                )

            groupby_names = resd["avg"].index.names
            groupby = "\t".join(groupby_names)
            obs_names = resd["avg"].index.map(lambda x: "\t".join(str(y) for y in x))
            obs = resd["obs"].copy()
            obs["cell_count"] = resd["ncells"]
            obs.index = obs_names
            adata.obs = obs
            adata.var_names = pd.Index(var_names, name="features")
            adata.obs_names = pd.Index(obs_names, name=groupby)

        adata.uns["approximation_groupby"] = {
            "names": groupby_names,
            "dtypes": groupby_dtypes,
        }
        if neighborhood:
            adata.uns["approximation_groupby"]["order"] = resd["avg"].index.values

        return adata

    def to_anndata(
        self,
        groupby="celltype",
        neighborhood=False,
        measurement_type=None,
    ):
        if measurement_type is None:
            with h5py.File(self.filename) as h5_data:
                measurement_type = self._infer_measurement_type(h5_data)

        if isinstance(groupby, str):
            groupby = [groupby]
        groupby = tuple(groupby)

        if (measurement_type, neighborhood) not in self._adata_dict:
            self._to_anndata(
                neighborhood=neighborhood,
                measurement_type=measurement_type,
            )

        # FIXME: specify that it's a view somehow
        adata = self._adata_dict[(measurement_type, neighborhood)]

        # Coarse grain result if the approximation includes unnecessary metadata
        adata_cg = coarse_grain_anndata(adata, groupby)
        return adata_cg



import pathlib
import yaml



def load_config(config_path):
    with open(config_path) as f:
        config = yaml.safe_load(f)

    # Propagate cell supertypes and order for simplicity
    if 'cell_annotations' in config:
        for mt in config['measurement_types']:
            config_mt = config[mt]
            if 'cell_annotations' not in config_mt:
                config_mt['cell_annotations'] = config['cell_annotations']
            else:
                for st in config['cell_annotations']:
                    config_mt['cell_annotations'][st] = config['cell_annotations'][st]

    for mt in config["measurement_types"]:
        config_mt = config[mt]

        if ("path" not in config_mt) and ("path_global" not in config_mt):
            config_mt["path"] = species + '.h5ad'

        if ("path" in config_mt) and isinstance(config_mt["path"], str):
            config_mt["path"] = {t: config_mt["path"] for t in config_mt["tissues"]}

        # Use absolute paths
        root_fdn = root_repo_folder / 'data' / 'full_atlases' / mt / species
        for key in ["path_global", "path_metadata_global"]:
            if key in config_mt:
                config_mt[key] = root_fdn / config_mt[key]
        if "path" in config_mt:
            for tissue in config_mt["path"]:
                config_mt["path"][tissue] = root_fdn / config_mt["path"][tissue]

        if "filter_cells" not in config_mt:
            config_mt["filter_cells"] = {}

        if "require_subannotation" not in config_mt["cell_annotations"]:
            config_mt["cell_annotations"]["require_subannotation"] = []

        if "subannotation_kwargs" not in config_mt["cell_annotations"]:
            config_mt["cell_annotations"]["subannotation_kwargs"] = {}

        if "blacklist" not in config_mt["cell_annotations"]:
            config_mt["cell_annotations"]["blacklist"] = {}

        celltype_order = []
        for supertype in config_mt["cell_annotations"]["supertype_order"]:
            ct_order_supertype = (
                supertype, config_mt["cell_annotations"]["cell_supertypes"][supertype],
            )
            celltype_order.append(ct_order_supertype)
        config_mt["cell_annotations"]["celltype_order"] = celltype_order

        del config_mt["cell_annotations"]["supertype_order"]
        del config_mt["cell_annotations"]["cell_supertypes"]

        if "feature_annotation" not in config_mt:
            config_mt["feature_annotation"] = False
        else:
            # FIXME
            config_mt["feature_annotation"] = root_repo_folder / 'data' / 'gene_annotations' / config_mt["feature_annotation"]

    return config





import numpy as np
import anndata


def coarse_grain_anndata(adata, groupby):
    """Coarse grain an approximation further to broader grouping."""
    if 'approximation_groupby' not in adata.uns:
        raise KeyError(
            "missing .uns['approximation_groupby'] information, this does not look like an approximation",
        )

    if isinstance(groupby, str):
        groupby = [groupby]

    groupby = list(groupby)
    if len(groupby) == 0:
        raise ValueError(
            'groupby must be a sequence with at least one element',
        )

    groupby_original = list(adata.uns['approximation_groupby']['names'])
    for column in groupby:
        if column not in groupby_original:
            raise ValueError(
                f"Grouping not found in approximation: {column}",
            )

    # Trivial case
    if groupby_original == groupby:
        return adata.copy()

    indices_groupby = [groupby_original.index(gb) for gb in groupby]

    # Detect neighborhood
    neighborhood = 'X_ncells' in adata.obsm
    if neighborhood:
        # We only have to touch the X_ncells, the rest is grouping agnostic
        multiindex = (pd.Index(adata.uns['approximation_groupby']['order'])
                        .str
                        .split('\t', expand=True))
        multiindex.names = groupby_names
        ncells = pd.DataFrame(
            adata.obsm['X_ncells'],
            index=multiindex,
        )
        # FIXME: this probably needs more work
        ncells_cg = ncells.groupby(level=indices_groupby).sum()
        new_order = np.asarray(
            ['\t'.join(x) for x in ncells_cg.index],
        )

        adata_cg = adata.copy()
        adata_cg.obsm['X_ncells'] = ncells_cs.values
        adata_cg.uns['approximation_groupby'] = {
            'names': groupby,
            'dtypes': [adata.uns['approximation_groupby']['dtypes'][i] for i in indices_groupby],
            'order': new_order,
        }

    else:
        # If not neighborhood, we have to actually change the matrix X and, if present,
        # the layers (e.g. avg and fractions)
        ncells_with_meta = adata.obs[list(groupby_original) + ['cell_count']].copy()
        ncells_with_meta['idx'] = np.arange(len(ncells_with_meta))

        # To coarse grain, we need two things:
        # 1. the number of cells in each group
        # 2. the index of cells in each group
        gby = ncells_with_meta.groupby(groupby)

        X = np.zeros((gby.ngroups, adata.X.shape[1]), adata.X.dtype)
        obs_names = []
        if adata.layers:
            layers = {
                key: np.zeros((gby.ngroups, layer.shape[1]), layer.dtype) for key, layer in adata.layers.items()
            }

        for i, (groupid, group) in enumerate(gby):
            idx = group['idx'].values
            ncell = group['cell_count'].values
            ncell_cg = ncell.sum()
            #print(groupid)
            #print(group)
            #print(adata.X[idx].shape)
            X[i] = (adata.X[idx] * ncell).sum(axis=0) / ncell_cg
            if adata.layers:
                for key in layers:
                    layers[key][i] = (adata.layers[key][idx] * ncell).sum(axis=0) / ncell_cg
            obs_name = '\t'.join(str(gid) for gid in groupid)
            obs_names.append(obs_name)

        if adata.layers:
            adata_cg = anndata.AnnData(
                X=X,
                var=adata.var.copy(),
            )
        else:
            adata_cg = anndata.AnnData(
                X=X,
                layers=layers,
                var=adata.var.copy(),
            )

        # Obs names and metadata
        ncells_cg = gby.sum()['cell_count']
        adata_cg.obs_names = obs_names
        adata_cg.obs['cell_count'] = ncells_cg.values
        for gb in groupby:
            adata_cg.obs[gb] = ncells_cg.index.get_level_values(gb)

        adata_cg.uns['approximation_groupby'] = {
            'names': groupby,
            'dtypes': [adata.uns['approximation_groupby']['dtypes'][i] for i in indices_groupby],
        }

    return adata_cg




import numpy as np
import matplotlib.pyplot as plt


def neighborhood_composition(
    accessor,
    groupby='celltype',
    measurement_type=None,
    ax=None,
    cmap='viridis_r',
    **kwargs,
):
    """Composition of all neighborhoods."""

    if ax is None:
        ax = plt.gca()

    adata = accessor.to_adata(
        groupby=groupby,
        neighborhood=True,
        measurement_type=measurement_type,
    )

    Xncells = adata.obsm['X_ncells']
    celltypes = adata.uns['celltypes']
    
    nnei, nct = Xncells.shape

    ax.set_xlim(-0.5, nct - 0.5)
    ax.set_ylim(-0.5, nnei - 0.5)
    xs, ys = np.meshgrid(np.arange(nct), np.arange(nnei))
    hs = 1.0 * Xncells / Xncells.max()

    facecolors = plt.cm.get_cmap(cmap)(hs.ravel())
    ax.bar(
        x=xs.ravel(),
        height=hs.ravel(),
        bottom=(ys - hs / 2).ravel(),
        color=facecolors,
        edgecolor=tuple([0.1] * 3 + [0.5]),
    )
    ax.set_xticks(np.arange(nct))
    ax.set_xticklabels(celltypes, rotation=90)
    ax.set_yticks(np.arange(nnei))
    ax.set_ylabel('Neighborhood')
    ax.figure.tight_layout()

    return ax



"""Compress single cell data using biological cell metadata groups, including cell type."""

import gc
import numpy as np
import pandas as pd
import scanpy as sc


def _compress_biological(
    adata,
    celltype_column,
    additional_groupby_columns,
    measurement_type,
):
    """Compress at the cell type level"""
    features = adata.var_names
    nfeatures = len(features)

    # Metadata for groups
    groupby_columns = [celltype_column] + list(additional_groupby_columns)
    tmp = adata.obs[groupby_columns].copy()
    tmp['cell_count'] = 1
    obs = tmp.groupby(groupby_columns, observed=False).sum().reset_index()
    ngroups = len(obs)

    # Names of rows
    obs_names = ['\t'.join(str(y) for y in x) for _, x in obs[groupby_columns].iterrows()]
    obs_names = pd.Index(
        obs_names,
        name='\t'.join(groupby_columns),
    )

    # Molecular counts
    avg = np.zeros((nfeatures, ngroups), np.float32)
    if measurement_type == "gene_expression":
        frac = np.zeros_like(avg)
    idx = np.zeros(adata.n_obs, bool)
    for i, row in obs.iterrows():
        # Reconstruct indices of focal cells
        idx[:] = True
        for col in groupby_columns:
            idx &= (adata.obs[col] == row[col])

        # Average across group
        Xidx = adata[idx].X
        avg[:, i] = np.asarray(Xidx.mean(axis=0))[0]
        if measurement_type == "gene_expression":
            frac[:, i] = np.asarray((Xidx > 0).mean(axis=0))[0]

    res = {
        'obs': obs,
        'obs_names': obs_names,
        'Xave': avg,
    }
    if measurement_type == "gene_expression":
        res['Xfrac'] = frac
    return res


